# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Registry for layers and their parameters/variables.

This represents the collection of all layers in the approximate Fisher
information matrix to which a particular FisherBlock may belong. That is, we
might have several layer collections for one TF graph (if we have multiple K-FAC
optimizers being used, for example.)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=unused-import,line-too-long,wildcard-import
from tensorflow.contrib.kfac.python.ops.layer_collection import *
from tensorflow.python.util.all_util import remove_undocumented
# pylint: enable=unused-import,line-too-long,wildcard-import

_allowed_symbols = [
    "get_default_layer_collection",
    "set_default_layer_collection",
    "LayerParametersDict",
    "LayerCollection",
    "APPROX_KRONECKER_NAME",
    "APPROX_DIAGONAL_NAME",
    "APPROX_FULL_NAME",
    "VARIABLE_SCOPE",
    "APPROX_KRONECKER_INDEP_NAME",
    "APPROX_KRONECKER_SERIES_1_NAME",
    "APPROX_KRONECKER_SERIES_2_NAME"
]

remove_undocumented(__name__, allowed_exception_list=_allowed_symbols)
